/*

	Sphelar + DLCAP 񃂃j^

	ATtiny13A

	Copyrights 2010,  Yoshi Hiroshi.

    FUSE Settings:
        CLOCK: 128KHz

 */

#include <avr/io.h>
#include <avr/sleep.h>
#include <avr/wdt.h>
#include <avr/interrupt.h>
#include <util/delay.h>

#define REF_VOLT                    1100L
#define LIMIT_VOLT                  2350L
#define TOTAL_VOLT_HIGHER           3200L

#define REGISTER_HIGH               47L  // 4.7M orm
#define REGISTER_LOW                20L  // 2.0M ohm
#define GET_CAP_VOLT(adc)           (((long)(adc) * REGISTER_HIGH * REF_VOLT) / (REGISTER_LOW * 1024L))

#define BYPASS_1_ON                 {PORTB &= ~0x02;}
#define BYPASS_1_OFF                {PORTB |= 0x02;}
#define BYPASS_2_ON                 {PORTB |= 0x10;}
#define BYPASS_2_OFF                {PORTB &= ~0x10;}
#define VOLT_HIGH                   {PORTB |= 0x01;}
#define VOLT_LOW                    {PORTB &= ~0x01;}

volatile int adcVolt1;
volatile int adcVolt2;

void adcEnable()
{
    ADCSRA = _BV(ADEN) | _BV(ADPS1);
    ADMUX = _BV(REFS0);
}

inline void adcDisable()
{
    ADMUX = 0;
    ADCSRA = 0;
}

void wdtEnable()
{
    MCUSR &= ~_BV(WDRF);
    WDTCR |= _BV(WDCE) | _BV(WDE);
    WDTCR = _BV(WDTIE) | _BV(WDP3) | _BV(WDP0);     // 8S;
}

void bypassCurrent1(int capVolt)
{
    if(capVolt >= LIMIT_VOLT) {
        BYPASS_1_ON;
    } else {
        BYPASS_1_OFF;
    }
}

void bypassCurrent2(int capVolt)
{
    if(capVolt >= LIMIT_VOLT) {
        BYPASS_2_ON;
    } else {
        BYPASS_2_OFF;
    }
}

ISR(WDT_vect)
{
}

int main()
{
    CLKPR = _BV(CLKPCE);
    CLKPR = 0;          // 128KHz
    PORTB = 0x02;
    DDRB  = 0x13;
    DIDR0 = 0x3E;
    wdtEnable();

    while(1)
    {
        set_sleep_mode(SLEEP_MODE_PWR_DOWN);
        sleep_enable();
        sei();
        sleep_cpu();
        sleep_disable();

        adcEnable();                            // ADCN
        _delay_us(200);
        ADMUX = _BV(REFS0) | 0x01;              // LpV^d1擾
        ADCSRA |= _BV(ADSC);
        loop_until_bit_is_set(ADCSRA, ADIF);    // ŏ̕ϊf[^j
        ADCSRA |= _BV(ADIF);
        ADCSRA |= _BV(ADSC);
        loop_until_bit_is_set(ADCSRA, ADIF);    // ϊ̏I҂
        ADCSRA |= _BV(ADIF);
        adcVolt1 = ADC;
        ADMUX = _BV(REFS0) | 0x03;              // LpV^d2擾
        ADCSRA |= _BV(ADSC);
        loop_until_bit_is_set(ADCSRA, ADIF);    // ϊ̏I҂
        ADCSRA |= _BV(ADIF);
        adcVolt2 = ADC;
        adcDisable();

        adcVolt1 = GET_CAP_VOLT(adcVolt1);
        bypassCurrent1(adcVolt1);                // dɂ[ddoCpX
        adcVolt2 = GET_CAP_VOLT(adcVolt2);
        bypassCurrent2(adcVolt2);                // dɂ[ddoCpX

        if((adcVolt1 + adcVolt2) >=  TOTAL_VOLT_HIGHER) {
            VOLT_HIGH;
        } else {
            VOLT_LOW;
        }
    }
}

