/*

	Sphelar + DLCAP ~d

	ATtiny45V

	Copyrights 2010,  Yoshi Hiroshi.

    FUSE Settings:
        CLOCK: 128KHz

 */

#include <avr/io.h>
#include <avr/sleep.h>
#include <avr/wdt.h>
#include <avr/interrupt.h>
#include <util/delay.h>

#define POWER_ON                    PORTB &= ~_BV(PB0)
#define POWER_OFF                   PORTB |= _BV(PB0)
#define IS_VOUT_LOW                 (ACSR & _BV(ACO))

#define IS_CURRENT_HIGH             (PINB & _BV(PB3))
#define CURRENT_MODE_LOW            1
#define CURRENT_MODE_HIGH           0xFF
#define IS_INVOLT_HIGH              (PINB & _BV(PB4))
#define INPUT_RANGE_UNDER           22
#define INPUT_RANGE_LOW             33
#define INPUT_RANGE_HIGH            0xFF

volatile unsigned char startingUp;
volatile unsigned char sleepMode;
volatile unsigned char currentMode;
volatile unsigned char inputRange;
volatile char loopCount;

void wdtEnable(char time)
{
    char wdp;
    switch(time) {
    case 0:
        wdp = _BV(WDTIE) | 0;                       // 16mS
        break;
    default:
        wdp = _BV(WDTIE) | _BV(WDP3) | _BV(WDP0);   // 8S
    }
    MCUSR &= ~_BV(WDRF);
    WDTCR = _BV(WDCE) | _BV(WDE);
    WDTCR = wdp;
}

ISR(PCINT0_vect)
{
}

ISR(WDT_vect)
{
}

int main()
{
    CLKPR = _BV(CLKPCE);
    CLKPR = 0;              // 128KHz
    PORTB = 0x01;
    DDRB |= 0x01;
    PRR = 0x0F;
    POWER_OFF;
    startingUp = 1;
    DIDR0 = _BV(ADC1D) | _BV(AIN1D);
    sleepMode = SLEEP_MODE_PWR_DOWN;
    wdtEnable(8);

    while(1)
    {
        if(sleepMode == SLEEP_MODE_PWR_DOWN) {
            set_sleep_mode(sleepMode);
            sleep_enable();
            sei();
            sleep_cpu();
            sleep_disable();
        }

        inputRange = IS_INVOLT_HIGH ? INPUT_RANGE_HIGH : INPUT_RANGE_LOW;
        if(startingUp != 0) {
            if(IS_INVOLT_HIGH) startingUp = 0;
        }
        if(startingUp == 0) {
            PRR &= ~_BV(PRADC);
            ACSR = _BV(ACBG);
            ADCSRB = _BV(ACME);
            ADMUX = 0x01;
            _delay_us(10);
            if(IS_VOUT_LOW) {                   // Vin2.6VȉH
                POWER_OFF;                      // o̓gWX^OFF
                startingUp = 1;
                PRR |= _BV(PRADC);
                ACSR = _BV(ACD);
                ADCSRB = 0;
                wdtEnable(8);                       // Ď8b
                PCMSK &= ~_BV(PCINT3);              // d[h̕ω݋֎~
                sleepMode = SLEEP_MODE_PWR_DOWN;
            } else {
                if(inputRange == INPUT_RANGE_LOW) { // ͓dႢƂ
                    //if(IS_VOUT_LOW == 0) 
                    {
                        POWER_ON;                       // o̓gWX^ON
                    }
                    PRR |= _BV(PRADC);
                    ACSR = _BV(ACD);
                    ADCSRB = 0;
                    wdtEnable(8);                       // Ď8b
                    PCMSK &= ~_BV(PCINT3);              // d[h̕ω݋֎~
                    sleepMode = SLEEP_MODE_PWR_DOWN;
                } else {                            // ͓dƂA
                    ADCSRB = 0;
                    currentMode = IS_CURRENT_HIGH ? CURRENT_MODE_HIGH : CURRENT_MODE_LOW;
                    if(currentMode == CURRENT_MODE_LOW) {   // dȂA
                        PRR |= _BV(PRADC);
                        sleepMode = SLEEP_MODE_PWR_DOWN;
                        loopCount = 0;
                        while(IS_VOUT_LOW) {                    // o͓dグāA
                            POWER_ON;
                            if(loopCount++ >= 560) break;
                        }
                        POWER_OFF;                              // TrOFF
                        wdtEnable(0);                           // Ď𑁂߂
                        PCMSK |= _BV(PCINT3);                   // d[hω݋
                    } else {                                // dȂA
                        sleepMode = SLEEP_MODE_IDLE;            // AĎ[hɂ
                        if(IS_VOUT_LOW) {
                            POWER_ON;                           // o͂ႢƂATrON
                        } else {
                            POWER_OFF;                          // o͂ƂATrOFF
                        }
                        PCMSK &= ~_BV(PCINT3);                  // d[hω݋֎~
                    }
                }
            }
        }
    }
}

