/*

                WetherMeter

                Cۏ̎WƋL^

	i2cTinyRtc.c

	IICڑRTCW[Rg[B

	Copyrights 2010,  Yoshi Hiroshi.

 */

#include "i2cTinyRtc.h"

unsigned char dmaxList[] = 
{
    31,28,31,30,31,30,31,31,30,31,30,31
};

unsigned char rtcBuff[9];

//------------------------------
unsigned char* getRtcBuff()
{
    return rtcBuff;
}

//------------------------------
void rtcDateRead()
{
    rtcRead(RTC_ADDRESS, 2, rtcBuff, 7);
    rtcBuff[0] &= 0x7F;
    rtcBuff[1] &= 0x7F;
    rtcBuff[2] &= 0x3F;
    rtcBuff[3] &= 0x3F;
    rtcBuff[4] &= 0x07;
    rtcBuff[5] &= 0x1F;
}

//------------------------------
void rtcDateWrite(unsigned char yy, unsigned char mm, unsigned char dd)
{
    unsigned char dcount;
    if(yy > 99) yy = 99;
    rtcBuff[8] = GET_HEX_DIGIT(yy);
    if(mm > 12) mm = 12;
    if(mm <  1) mm = 1;
    rtcBuff[7] = GET_HEX_DIGIT(mm);
    dcount = dmaxList[mm - 1];
    if(mm == 2) {
        if((yy % 4) == 0) dcount = 29;
    }
    if(dd < 1) dd = 1;
    if(dd > dcount) dd = dcount;
    rtcBuff[5] = GET_HEX_DIGIT(dd);
    dcount = (yy > 1)? (365 * (yy - 1) + ((yy - 1) / 4)) : 0;
    for(unsigned char mx = 0; mx < (mm - 1); mx++) {
        dcount += dmaxList[mx];
    }
    if((mm > 2) && ((yy % 4) == 0)) {
        dcount++;
    }
    dcount += dd - 1;
    rtcBuff[6] = (dcount + 6) % 7;

    rtcBuff[0] = 0x20;
    rtcWrite(RTC_ADDRESS, 0, rtcBuff, 1);   // stop rtc

    rtcWrite(RTC_ADDRESS, 5, &rtcBuff[5], 4);   // update  date

    rtcBuff[0] = 0x00;
    rtcWrite(RTC_ADDRESS, 0, rtcBuff, 1);   // restart rtc
}

//------------------------------
void rtcTimeWrite(unsigned char hh, unsigned char min, unsigned char ss)
{
    rtcBuff[4] = GET_HEX_DIGIT((hh > 23)? 23 : hh);
    rtcBuff[3] = GET_HEX_DIGIT((min > 59)? 59 : min);
    rtcBuff[2] = GET_HEX_DIGIT((ss > 59)? 59 : ss);
    rtcBuff[1] = 0;
    rtcBuff[0] = 0x20;
    rtcWrite(RTC_ADDRESS, 0, rtcBuff, 1);   // stop rtc
    rtcWrite(RTC_ADDRESS, 2, &rtcBuff[2], 3);   // set time
    rtcBuff[0] = 0x00;
    rtcWrite(RTC_ADDRESS, 0, rtcBuff, 1);   // restart rtc
}

//------------------------------
void rtcInit()
{
	RTC_SCL_FREE;
	RTC_SDA_FREE;
    rtcBuff[0] = 0;
    rtcBuff[1] = 0;
    rtcWrite(RTC_ADDRESS, 0x00, rtcBuff, 2);    // Control Reg. Init
    rtcBuff[0] = 0x83;
    rtcWrite(RTC_ADDRESS, 0x0D, rtcBuff, 1);    // Clockout -> 1Hz
    rtcRead(RTC_ADDRESS, 2, rtcBuff, 1);
    if(rtcBuff[0] & 0x80) 
    {
        rtcDateWrite(10, 1, 1);
        rtcTimeWrite(0, 0, 0);
    }
}

//------------------------------
void rtcSendStartCondition()
{
	RTC_SCL_FREE;			// free SCL line
	while( RTC_SCL_CURR_VAL == 0 );
	RTC_DELAY_LONG;
	RTC_SDA_SET_LOW;		// Initiate SDA line
	RTC_DELAY_LONG;
	RTC_SCL_SET_LOW;		// Initiate SCL line
}

//------------------------------
void rtcSendStopCondition()
{
	RTC_SDA_SET_LOW;
	RTC_DELAY_SHORT;
	RTC_SCL_FREE;
	while( RTC_SCL_CURR_VAL == 0 );
	RTC_DELAY_SHORT;
	RTC_SDA_FREE;
	RTC_DELAY_LONG;
}

//------------------------------
unsigned char rtcWriteData(unsigned char val)
{
	unsigned char result;
	for(unsigned char count = 0; count < 8; count++)
    {
        if(val & 0x80) {
        	RTC_SDA_FREE;			// send 1
        } else {
        	RTC_SDA_SET_LOW;	    // send 0
        }
		RTC_DELAY_LONG;
		RTC_SCL_FREE;               // toggle SCL to high(free)
		while( RTC_SCL_CURR_VAL == 0 );
		RTC_DELAY_LONG;
		RTC_SCL_SET_LOW;            // toggle SCL to low
        val = val << 1;
	}
    RTC_SDA_FREE;
	RTC_DELAY_LONG;
    RTC_SCL_FREE;               // toggle SCL to high(free)
	RTC_DELAY_LONG;
    result = RTC_SDA_CURR_VAL;
	RTC_SCL_SET_LOW;            // toggle SCL to low
	return (result == 0);
}

//------------------------------
unsigned char rtcWrite(unsigned char slaveAdrs, unsigned char control, unsigned char* data, unsigned char len)
{
    unsigned char count;
// o Send start
	rtcSendStartCondition();
// o Send Slave address & write flag
	if(rtcWriteData(((slaveAdrs << 1) | RTC_REQUEST_WRITE)))
	{
// o Send Control
		if(rtcWriteData(control))
		{
// o Send Data
		    for(count = 0;count < len; count++) {
                if(!rtcWriteData(*data++)) break;
            }
		}
	}
// o Send Stop
	rtcSendStopCondition();
//
	return 0;
}

//------------------------------
unsigned char rtcReadData(unsigned char ack)
{
	unsigned char result = 0;
	for(unsigned char count = 0; count < 8; count++)
    {
		RTC_SCL_FREE;               // toggle SCL to high(free)
		while( RTC_SCL_CURR_VAL == 0 );
        result = result << 1;
        result |= RTC_SDA_CURR_VAL? 1:0;
		RTC_SCL_SET_LOW;            // toggle SCL to low
	}
    if(ack) {
        RTC_SDA_FREE;               // send ack bit
    } else {
        RTC_SDA_SET_LOW;            // send ack bit
    }
	RTC_DELAY_LONG;
    RTC_SCL_FREE;               // toggle SCL to high(free)
	RTC_DELAY_LONG;
	RTC_SCL_SET_LOW;            // toggle SCL to low
    RTC_SDA_FREE;               // release data bit
	return result;
}

//------------------------------
unsigned char rtcRead(unsigned char slaveAdrs, unsigned char control, unsigned char* data, unsigned char len)
{
    unsigned char count = 0;
// o Send start
	rtcSendStartCondition();
// o Send Slave address & read/write flga
	if(rtcWriteData(((slaveAdrs << 1) | RTC_REQUEST_WRITE)))
	{
// o Send Control
		if(rtcWriteData(control))
		{
// o Send start
        	rtcSendStartCondition();
// o Send Slave address & read flag
        	if(rtcWriteData(((slaveAdrs << 1) | RTC_REQUEST_READ)))
        	{
    		    for(count = 0; count < len; count++) {
                    *data ++ = rtcReadData(count == (len - 1));
                }
            }
        }
	}
// o Send Stop
	rtcSendStopCondition();
//
	return count;
}

