/*
                WetherMeter

                Cۏ̎WƋL^

			Copyrights 2010,  Yoshi Hiroshi

            WetherMeter:   Sparkfun
                ́AAJ
                PC0:J 0.294mm / pulse
                PC1: 2.4Kmh / pulse/sec
                PC2: 

            CZT SCP1000-D01FVTI
                        BreakoutFSparkfun
            xZT SHT-11FSensirion

        AVR:ATmega88
        FUSE: CLOCK8MHz, CHKDIV81/8
              0x42, 0xD7, 0x03
 */

#include <avr/io.h>
#include <avr/sleep.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <util/delay.h>
#include <ctype.h>
#include "uart.h"
#include "i2cTinyRtc.h"
#include "i2cTinySht.h"
#include "spiSCP1000.h"

#define RAIN_SWITCH         (PINC & _BV(PC0))
#define WIND_SWITCH         (PINC & _BV(PC1))

#define POWER_LARGE         (PORTD |= _BV(PD7))
#define POWER_SMALL         (PORTD &= ~_BV(PD7))
#define COM_SLEEP           (PORTD |= _BV(PD2))
#define COM_AWAKE           (PORTD &= ~_BV(PD2))

volatile char timerIntFlag;
volatile char timerState;
volatile unsigned char timerHour;
volatile unsigned char timerMinute;
volatile unsigned char timerSecond;
volatile unsigned char currentMinute;
volatile unsigned char currentSecond;
volatile char startingUpFlag;
volatile char timeModReq;
volatile int startupTimer;
volatile unsigned char sleepMode;
volatile unsigned char forcedSend;

struct histData_ {
    int temp;
    int hum;
    int press;
    int anemo;
    char anemoMax;
    char vane;
    char rain;
} histData[30];
volatile struct histData_ *histPtr;
volatile int anemoCountSave;
volatile char rainCountSave;

volatile int anemoCount;
volatile char anemoSec;
volatile char anemoMax;
volatile char anemoState;
volatile int windVaneAdc;
volatile int windVane;
volatile unsigned char windVaneIndex;

volatile char rainCount;
volatile char rainState;

volatile long humidityVal;
volatile unsigned char pressureHigh;
volatile unsigned int pressureLow;
volatile int  temperatureVal;
volatile long pressureVal;

volatile char udreIntFlag;

struct windVaneTable_
{
    unsigned int dir;
    unsigned int voltage;
    unsigned char str[4];
} windVaneTable[] EEMEM =
{
    {  3150,                    0,  "NW"  },
    {  1125,   32L * 1024L / 500L,  "SEE" },
    {   675,   41L * 1024L / 500L,  "NEE" },
    {   900,   45L * 1024L / 500L,  "E"   },
    {  1575,   62L * 1024L / 500L,  "SSE" },
    {  1350,   90L * 1024L / 500L,  "SE"  },
    {  2025,  119L * 1024L / 500L,  "SSW" },
    {  1800,  140L * 1024L / 500L,  "S"   },
    {   225,  198L * 1024L / 500L,  "NNE" },
    {   450,  225L * 1024L / 500L,  "NE"  },
    {  2475,  293L * 1024L / 500L,  "SWW" },
    {  2250,  308L * 1024L / 500L,  "SW"  },
    {  3375,  343L * 1024L / 500L,  "NNW" },
    {     0,  384L * 1024L / 500L,  "N"   },
    {  2925,  404L * 1024L / 500L,  "NWW" },
    {  2700,  462L * 1024L / 500L,  "W"   },
    {  3150,  478L * 1024L / 500L,  "NW"  }
};

//==================================================
//
//      ADϊ̓JnBdB
//
//==================================================
void adcEnable()
{
    PRR &= ~_BV(PRADC);
    ADCSRA = _BV(ADEN);
    ADMUX = _BV(REFS0) | 2;     // Vref = Vcc
    PORTC |= _BV(PC3);  // p@vAbvR
}

//==================================================
//
//      ADϊ̓~BdȂB
//
//==================================================
void adcDisable()
{
    ADMUX = 0;
    ADCSRA = 0;
    PORTC &= ~_BV(PC3);  // p@vAbvR֎~
    PRR |= _BV(PRADC);
}

//==================================================
//
//      NbNɐݒ肷
//
//==================================================
void clockFast()
{
    CLKPR = _BV(CLKPCE);
    CLKPR = 3;              // 1MHz
}

//==================================================
//
//      NbNᑬɐݒ肷
//
//==================================================
void clockSlow()
{
    CLKPR = _BV(CLKPCE);
    CLKPR = 8;              // 31.25KHz
}

//==================================================
//
//      tA̕ύXsB
//
//          Dyymmdd
//          Thhmmss
//          G
//
//==================================================
char checkAndDoCommand(char cmnd)
{
    char result = 0;    // 1:R}h̊
    static char order = 0;
    static int p1,p2,p3;
    static char charCount;
    if(order == 0) {
        cmnd |= 0x20;
        if(cmnd == 'd') {
            p1 = 10;
            p2 = p3 = 1;
            charCount = 0;
            order = cmnd;
        } else if(cmnd == 't') {
            p1 = p2 = p3 = 0;
            charCount = 0;
            order = cmnd;
        } else if(cmnd == 'g') {
            result = 1;
            printf_P(PSTR("o!\n"));
        } else if(!isspace(cmnd)) {
            printf_P(PSTR("<ILL CMND:[%c]\n"), cmnd);
        }
    } else {
        if(isdigit(cmnd)) {
            switch(charCount) {
            case 0:
                p1 = (cmnd & 0x0F) * 10;
                break;
            case 1:
                p1 += (cmnd & 0x0F);
                break;
            case 2:
                p2 = (cmnd & 0x0F) * 10;
                break;
            case 3:
                p2 += (cmnd & 0x0F);
                break;
            case 4:
                p3 = (cmnd & 0x0F) * 10;
                break;
            case 5:
                p3 += (cmnd & 0x0F);
                break;
            }
            charCount++;
            if(charCount == 6) {
                if(order == 'd') {
                    printf_P(PSTR(" <= Date:%02d-%02d-%02d\n"), p1, p2, p3);
                    rtcDateWrite(p1, p2, p3);
                } else if(order == 't') {
                    printf_P(PSTR(" <= Time:%02d:%02d:%02d\n"), p1, p2, p3);
                    rtcTimeWrite(p1, p2, p3);
                }
                order = 0;
            }
        } else {
            printf_P(PSTR("<ILL DIGIT CHARA:[%c]\n"), cmnd);
            order = 0;
        }
    }
    return result;
}

//==================================================
//
//      qXg[f[^̏
//
//==================================================
void initHistData()
{
    for(unsigned char ix = 0; ix < 60; ix++)
    {
        histData[ix].hum = -1;
    }
}

//==================================================
//
//      ݓo͂
//
//==================================================
void printCurrDate()
{
    rtcDateRead();
    timerHour   = GET_DIGIT(getRtcBuff()[2]);
    timerMinute = GET_DIGIT(getRtcBuff()[1]);
    timerSecond = GET_DIGIT(getRtcBuff()[0]);
    printf_P(PSTR("20%02x/%02x/%02x %02x:%02x:%02x\n"), 
        getRtcBuff()[6],getRtcBuff()[5],getRtcBuff()[3],
        getRtcBuff()[2],getRtcBuff()[1],getRtcBuff()[0]);
}

//==================================================
//
//      N̏
//
//==================================================
unsigned char startingUp()
{
    unsigned char result = 0;
    if(startupTimer == 0) {
        sleepMode = SLEEP_MODE_IDLE;
        clockFast();
        POWER_LARGE;
        COM_AWAKE;
        PRR &= ~_BV(PRUSART0);
    }
    // XBee肷܂ő҂
    if(timeModReq == 0) {
        if(startupTimer == 10) printf_P(PSTR("1"));
        else if(startupTimer == 20) printf_P(PSTR("2"));
        else if(startupTimer == 30) printf_P(PSTR("3"));
        else if(startupTimer <= 29) printf_P(PSTR("."));
    }

    if(startupTimer == 31) {        // NbZ[W
        printf_P(PSTR("\nWetherMeter started at "));
        printCurrDate();
    }
    if(startupTimer == 32) {
        COM_SLEEP;
        clockSlow();
        PRR |= _BV(PRUSART0);
        sleepMode = SLEEP_MODE_PWR_DOWN;
        POWER_SMALL;
    }

    if((startupTimer >= 32) &&      // 58ȍ~͑M̂ŁA
        !(timerMinute >= 58) ) {    // Nƕss
        result = 1;
    }
    return result;
}

//==================================================
//
//
//
//==================================================
void doMeasure()
{
    shtStartMeassure();     // xvJn
    scpWrite(0x03, 0x0C);   // CvJn
    PCMSK2 |= _BV(PCINT20); // xv݋

    adcEnable();            // ADϊ퓮Jn
    ADCSRA |= _BV(ADSC);    // ϊJn
    loop_until_bit_is_set(ADCSRA, ADIF);    // ϊ҂
    ADCSRA |= _BV(ADIF);    // tONA
    ADCSRA |= _BV(ADSC);    // ϊJn
    loop_until_bit_is_set(ADCSRA, ADIF);    // ϊ҂
    ADCSRA |= _BV(ADIF);    // tONA
    windVaneAdc = ADC;      //  Rl̕ωdɊZ
    adcDisable();           //ADϊ퓮~Ad~
    if((windVaneAdc > 1) && (windVaneAdc < 0x3F0)) {    // d畗
        for(unsigned char ix = 0; ix < 16;ix++)
        {
            if(((eeprom_read_word(&windVaneTable[ix].voltage) + 
                 eeprom_read_word(&windVaneTable[ix + 1].voltage)) / 2) >= windVaneAdc) {
                windVaneIndex = ix;
                break;
            } else {
                windVaneIndex = ix + 1;
            }
        }
        windVane = (int)eeprom_read_word(&windVaneTable[windVaneIndex].dir);
    } else {
        windVane = 225 * 99;    // vڑ́A99ݒ
    }
}

//==================================================
//
//
//
//==================================================
void saveData(unsigned char dataIndex)
{
    histPtr = &histData[dataIndex];
    histPtr->temp = temperatureVal;
    histPtr->press = (pressureVal + 5) / 10;
    histPtr->hum = humidityVal;
    histPtr->anemo = anemoCountSave;
    histPtr->anemoMax = anemoMax;
    histPtr->vane = windVane / 225;
    histPtr->rain = rainCountSave;
    anemoMax = 0;
}

//==================================================
//
//      1̌ĂяoŁA10Ԃ̃f[^𑗏o
//
//==================================================
void sendResult(unsigned char timeIndex)
{
    volatile unsigned char hIndex = timeIndex * 10;
    volatile unsigned char hIndexLim = hIndex + 10;
    if(hIndexLim > 60) hIndexLim = 60;

    while(hIndex < hIndexLim) {
        histPtr = &histData[hIndex % 30];
        if(histPtr->hum > 0) 
        {
// 
            printf_P(PSTR("%02d:%02d"), 
                timerHour, 
                hIndex);
// ÁAJ
            printf_P(PSTR(",%d,%d,%d,%d"), 
                histPtr->vane,
                histPtr->anemo, 
                histPtr->anemoMax, 
                histPtr->rain);
// C
            if(histPtr->temp >= 0) {
                printf_P(PSTR(",%d.%02d"), 
                    (int)(histPtr->temp/100), 
                    (int)(histPtr->temp%100));
            } else {
                printf_P(PSTR(",-%d.%02d"), 
                    (int)((-histPtr->temp)/100), 
                    (int)((-histPtr->temp)%100));
            }
// x
            printf_P(PSTR(",%d.%02d"), 
                histPtr->hum/100,
                histPtr->hum%100);
// C
            printf_P(PSTR(",%d.%01d"), 
                (int)(histPtr->press/10), 
                (int)(histPtr->press%10));
            printf_P(PSTR("\n"));
        }
        hIndex++;
    }
}

//--------------------------------------------------
//
//      PB1:SCP1000 DRDY
//
//--------------------------------------------------
ISR(PCINT0_vect)
{
    if(PINB & _BV(PB1)) {       // SCP1000 v
        temperatureVal = scpReadTemperature();
        pressureVal = scpReadPressure();
    }
}

//--------------------------------------------------
//
//    PC0:    2.4Kmh / pulse/sec
//    PC1:J    0.294mm / pulse
//
//--------------------------------------------------
ISR(PCINT1_vect)
{
    unsigned char swWind = WIND_SWITCH;
    unsigned char swRain = RAIN_SWITCH;
    if((anemoState == 0) && (swWind != 0)) {
        anemoCount++;
        anemoSec++;
        anemoState = 1;
    } else if((anemoState != 0) && (swWind == 0)) {
        anemoState = 0;
    }

    if((rainState == 0) && (swRain != 0)) {
        rainCount++;
        rainState = 1;
    } else if((rainState != 0) && (swRain == 0)) {
        rainState = 0;
    }
}

//--------------------------------------------------
//
//      PD5: RTC CLKOUT
//      PD4: SHT11 SDA
//
//--------------------------------------------------
ISR(PCINT2_vect)
{
    unsigned char intFlag = PIND & (_BV(PD6) | _BV(PD5) | _BV(PD4));
    unsigned char sw = intFlag & _BV(PD5);
    if((timerState == 0) && (sw == 0)) {
        timerState = 1;
        timerIntFlag = 1;
        if(anemoMax < anemoSec) anemoMax = anemoSec;
        anemoSec = 0;
        timerSecond++;
        if(timerSecond >= 60) {
            timerSecond = 0;
            timerMinute++;
            if(timerMinute >= 60) {
                timerMinute = 0;
                timerHour++;
                if(timerHour >= 24) {
                    timerHour = 0;
                }
            }
            anemoCountSave = anemoCount;
            anemoCount = 0;
            rainCountSave = rainCount;
            rainCount = 0;
        }
    } else if((timerState != 0) && (sw != 0)) {
        timerState = 0;
    }

    if(PCMSK2 & _BV(PCINT20)) {
        if((intFlag & _BV(PD4)) == 0) {       // SHT11 v
            PCMSK2 &= ~_BV(PCINT20);
            humidityVal = shtReadWord();
            humidityVal = -400L + (405L * humidityVal) / 100L;
        }
    }

    if((intFlag & _BV(PD6)) == 0) {
        if(forcedSend == 0) forcedSend = 1;
    }
}

//--------------------------------------------------
//
//
//
//--------------------------------------------------
ISR(USART_UDRE_vect)
{
    udreIntFlag = 1;
}

//--------------------------------------------------
//
//
//
//--------------------------------------------------
void startingUpSupport()
{
    if(startingUp()) {
        startingUpFlag = 0;
        anemoCountSave = anemoCount = anemoMax = 0;
        rainCountSave = rainCount = 0;
    }
    if(timeModReq == 0) {
        startupTimer++;
    } else if(timeModReq == 2) {
        startupTimer = 31;
        timeModReq = 0;
    }
}

//--------------------------------------------------
//
//
//
//--------------------------------------------------
void uartSupport()
{
    uartCheckAndSend();
    if(uartCheckAndRecv()) {
        char newChar;
        if((newChar = getchar()) != 0) {
            if(timeModReq == 0) 
            {
                printf_P(PSTR("\nSetup mode at "));
                printCurrDate();
            }
            timeModReq = 1;             // in command mode
            printf_P(PSTR("%c"), newChar);      // echo back character
            if(newChar == '\r') printf_P(PSTR("\n"));
            if(checkAndDoCommand(newChar)) {
                timeModReq = 2;         // end of command mode
            }
        }
    }
}

//--------------------------------------------------
//
//
//
//--------------------------------------------------
void timerInterrupt()
{
    if(startingUpFlag) {
        startingUpSupport();
    } else {
        currentSecond = timerSecond;    // ݎێ
        currentMinute = timerMinute;
        if(currentSecond == 0) {    // f[^vtF[Y
            doMeasure();
        }
        if(currentSecond == 2) {    // vf[^̊i[tF[Y
            saveData(currentMinute % 30);
        }
//              f[^MFXBeẽANeBu
        if((((currentMinute % 30) == 28) || (forcedSend == 1)) 
            && (currentSecond == 20)) 
        {
            POWER_LARGE;
            COM_AWAKE;
            forcedSend = 2;
        }
//              f[^̑MF10f[^3bԂőM
        if(((currentMinute % 30) == 29) || (forcedSend == 2)) 
        {
            if((currentSecond >= 3) &&
               (currentSecond <= 5)) {
                clockFast();
                PRR &= ~_BV(PRUSART0);
                sleepMode = SLEEP_MODE_IDLE;
                sendResult((currentSecond - 3) + 3 * (currentMinute / 30) );
            }
//@                MAXBeeX[v
            if(currentSecond == 7) {
                COM_SLEEP;
                POWER_SMALL;
                PRR |= _BV(PRUSART0);
                sleepMode = SLEEP_MODE_PWR_DOWN;
                initHistData();
                clockSlow();
                forcedSend = 0;
            }
        }
     }
}

//--------------------------------------------------
//
//
//
//--------------------------------------------------
void init()
{
// PORT Setting
    DDRC = 0x08;
    DDRD = 0x86;    // TxD,POWER_LARGE,COM_AWAKE
    PORTD |= _BV(PD6);  // SendReq Switch pullup
// XBee Power control
    POWER_LARGE;
    COM_AWAKE;
// Init Variables
    startingUpFlag = 1;
    timeModReq = 0;
    startupTimer = 0;
    sleepMode = SLEEP_MODE_PWR_DOWN;
    timerState = 0;
    anemoState = 0;
    rainState = 0;
    initHistData();
// VAʐM
    initUART();
// CZT[̏
    scpInit();
// xv
    shtInit();
// RTC̏
    rtcInit();
// Power Reduction
    PRR = 0xEF;
// Pin Change Interrupt Settings
    PCMSK0 = _BV(PCINT1);   // PB1:enable SCP1000 data ready int.
    PCMSK1 = _BV(PCINT9) | _BV(PCINT8); // PC0,PC1:J,
    PCMSK2 = _BV(PCINT22) | _BV(PCINT21);  // PD5,PD6:1Hz clock, force send
    PCICR  = _BV(PCIE2) | _BV(PCIE1) | _BV(PCIE0);
}

//--------------------------------------------------
//
//
//
//--------------------------------------------------
int main()
{
// Clock Prescaler
    clockFast();
// 
    init();

    while(1)
    {
        if(sleepMode == SLEEP_MODE_PWR_DOWN) 
        {
            set_sleep_mode(sleepMode);
            sleep_enable();
            sei();
            sleep_cpu();
            sleep_disable();
        }
 
        if(timerIntFlag) {
            timerIntFlag = 0;
            timerInterrupt();
        }

        uartSupport();
    }
}
